// Copyright (c) 2009, 2010 Paul Pogonyshev.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.


#ifndef MCT_IMPL_CLOSED_HASH_TABLE_HPP
#define MCT_IMPL_CLOSED_HASH_TABLE_HPP


#include <cmath>
#include <cstddef>
#include <cstring>
#include <iterator>
#include <memory>
#include <stdexcept>
#include <utility>

#if MCT_CXX0X_SUPPORTED
# include <initializer_list>
#endif

#include <mct/impl/optimization.hpp>
#include <mct/impl/utils.hpp>


// Internally data is stored in an array of buckets with length being some power of 2.
// Tables come in two quite different variants: those where buckets keep hash values of
// their keys and those where buckets don't.
//
// Hashes are kept in buckets:
// - each bucket has an additional field '_hash';
// - empty buckets have '_hash' set to 0, debris [*] have it set to 1;
// - hashes are always 'postprocessed' after they are computed by the actual hash
//   function so that hashes in used buckets are never 0 or 1;
// - this mode can be used with any key type.
//
// Hashes are recomputed each time:
// - for empty/debris buckets 0 or 1 is stored where key would be (we have to use some
//   reinterpret_cast black magic for this);
// - at most two buckets are designated as 'special': those are the buckets that have
//   keys that match the marks above, yet contain real values; such buckets can be located
//   anywhere in the table or be not present at all;
// - can only be used for 'small' keys: at most sizeof (long long).
//
// Default is to use the second mode (smaller memory footprint) if key size permits.
//
// If you try to improve this remember that key equality is not the same as bitwise
// equality, especially for structures and other custom types.  Even for plain integral
// type one could specify a different comparator function.  Additionally, you can't call
// hash and comparison functions on non-inserted data: they could assert something or
// dereference an invalid pointer as a result.
//
// [*] Debris are buckets that contained some value but were subsequently erased.  With
//     quadratic probing such buckets cannot be just reset to empty state.

// It was decided that tables never shrink on their own.  See documentation for rationale.


#if MCT_CXX0X_SUPPORTED
# define MCT_STD_FORWARD(type, value)  std::forward <type> (value)
#else
# define MCT_STD_FORWARD(type, value)  value
#endif


namespace mct
{

  namespace impl
  {

    template <std::size_t key_type_size>
    struct key_storage
    {
      static  bool
      is_empty (const void* storage)
      {
        return (is_zero_bytes_start (storage)
                && *(static_cast <const unsigned char*> (storage) + (key_type_size) - 1) == 0);
      }

      static  bool
      is_debris (const void* storage)
      {
        return (is_zero_bytes_start (storage)
                && *(static_cast <const unsigned char*> (storage) + (key_type_size) - 1) == 1);
      }

      static  bool
      is_empty_or_debris (const void* storage)
      {
        return (is_zero_bytes_start (storage)
                && *(static_cast <const unsigned char*> (storage) + (key_type_size) - 1) <= 1);
      }

      static  void
      clear (void* storage)
      {
        zero_start (storage);
        *(static_cast <unsigned char*> (storage) + (key_type_size) - 1) = 0;
      }

      static  void
      destroy (void* storage)
      {
        zero_start (storage);
        *(static_cast <unsigned char*> (storage) + (key_type_size) - 1) = 1;
      }

    private:

      static  bool
      is_zero_bytes_start (const void* storage)
      {
        for (const unsigned char* first = static_cast <const unsigned char*> (storage),
                                * last  = (first + (key_type_size) - 1);
             first < last; ++first)
          {
            if (*first)
              return false;
          }

        return true;
      }

      static  void
      zero_start (void* storage)
      {
        std::memset (storage, 0, (key_type_size) - 1);
      }
    };

    // Apparently, this is the standard and correct way to avoid pointer aliasing
    // problems.  Since we've already had such problems on GCC 4.5, let's be extra-safe.
    // A previous solution/workaround used converting 'storage' to 'volatile
    // integral_type*', but that's not correct according to the standard, even if works on
    // GCC 4.5.
    //
    // Hopefully, compilers are able to understand what's going on and optimize all the
    // complications away.  At least GCC seems to be able to do this, although
    // benchmarks do run a little bit slower.

# define MCT_SPECIALIZE_KEY_STORAGE(num_bytes, integral_type)   \
    template <>                                                 \
    struct key_storage <num_bytes>                              \
    {                                                           \
      static  bool                                              \
      is_empty (const void* storage)                            \
      {                                                         \
        integral_type  key;                                     \
        std::memcpy (&key, storage, sizeof (key));              \
        return key == 0;                                        \
      }                                                         \
      static  bool                                              \
      is_debris (const void* storage)                           \
      {                                                         \
        integral_type  key;                                     \
        std::memcpy (&key, storage, sizeof (key));              \
        return key == 1;                                        \
      }                                                         \
      static  bool                                              \
      is_empty_or_debris (const void* storage)                  \
      {                                                         \
        integral_type  key;                                     \
        std::memcpy (&key, storage, sizeof (key));              \
        return key <= 1;                                        \
      }                                                         \
      static  void                                              \
      clear (void* storage)                                     \
      {                                                         \
        const integral_type  key = 0;                           \
        std::memcpy (storage, &key, sizeof (key));              \
      }                                                         \
      static  void                                              \
      destroy (void* storage)                                   \
      {                                                         \
        const integral_type  key = 1;                           \
        std::memcpy (storage, &key, sizeof (key));              \
      }                                                         \
    }

    // What we do below with sizeof() trickery could just use 'uintN_t' types from
    // <stdint.h> or <cstdint>.  Unfortunately, those headers are not present on all
    // compilers.

    MCT_SPECIALIZE_KEY_STORAGE (MCT_SIZEOF_CHAR, unsigned char);

# if MCT_SIZEOF_SHORT > MCT_SIZEOF_CHAR
    MCT_SPECIALIZE_KEY_STORAGE (MCT_SIZEOF_SHORT, unsigned short);
# endif

# if MCT_SIZEOF_INT > MCT_SIZEOF_SHORT
    MCT_SPECIALIZE_KEY_STORAGE (MCT_SIZEOF_INT, unsigned int);
# endif

# if MCT_SIZEOF_LONG > MCT_SIZEOF_INT
    MCT_SPECIALIZE_KEY_STORAGE (MCT_SIZEOF_LONG, unsigned long);
# endif

# if MCT_HAVE_LONG_LONG && MCT_SIZEOF_LONG_LONG > MCT_SIZEOF_LONG
    MCT_SPECIALIZE_KEY_STORAGE (MCT_SIZEOF_LONG_LONG, unsigned long long);
# endif

# undef MCT_SPECIALIZE_TYPE_PICKERS


    template <typename Traits, bool keep_hashes>
    class bucket_base;


    template <typename Traits>
    class bucket_base <Traits, false>
    {
    public:

      typedef  typename Traits::value_type             value_type;
      typedef  typename Traits::assignable_value_type  assignable_value_type;
      typedef  typename Traits::key_type               key_type;

      static  const bool  KEEPS_HASHES             = false;
      static  const bool  SPECIAL_EMPTY_AND_DEBRIS = true;
      static  const bool  USES_SENTINEL            = false;
      static  const bool  ITERATORS_NEED_DATA      = true;


    private:

      value_type  _value;


    public:

      static  const key_type&
      extract_key (const value_type& value)
      {  return Traits::extract_key (value);  }

      const key_type&
      key () const
      {  return Traits::extract_key (_value);  }


      value_type&
      value ()
      {  return _value;  }

      const value_type&
      value () const
      {  return _value;  }

      template <typename Allocator>
      void
      construct_value (Allocator& allocator, const value_type& initializer)
      {  allocator.construct (&_value, initializer);  }

#   if MCT_CXX0X_SUPPORTED

      template <typename Allocator>
      void
      construct_value (Allocator& allocator, value_type&& initializer)
      {  allocator.construct (&_value, std::forward <value_type> (initializer));  }

#   endif


      template <typename Hash>
      std::size_t
      key_hash (const Hash& hasher) const
      {
        return hasher (Traits::extract_key (_value));
      }

      static  std::size_t
      postprocess_hash (std::size_t hash)
      {  return hash;  }


      bool
      used_and_matches (std::size_t hash) const
      {  return !is_empty_or_debris ();  }


      bool
      is_empty () const
      {
        return key_storage <sizeof (key_type)>::is_empty (static_cast <const void*> (&_value));
      }

      bool
      is_debris () const
      {
        return key_storage <sizeof (key_type)>::is_debris (static_cast <const void*> (&_value));
      }

      bool
      is_empty_or_debris () const
      {
        return key_storage <sizeof (key_type)>::is_empty_or_debris (static_cast <const void*>
                                                                    (&_value));
      }


      void
      store_hash (std::size_t hash)
      { }

      void
      copy_hash (const bucket_base& that)
      { }


      // See the comment above is_empty().

      void
      clear ()
      {
        return key_storage <sizeof (key_type)>::clear (static_cast <void*> (&_value));
      }

      void
      destroy ()
      {
        return key_storage <sizeof (key_type)>::destroy (static_cast <void*> (&_value));
      }


      void
      revert_to_empty ()
      {  clear ();  }

      void
      revert_to_debris ()
      {  destroy ();  }


      template <typename Data>
      static  typename Data::bucket_type*
      first (const Data& data)
      {
        typedef  typename Data::bucket_type  bucket_type;

        bucket_type* const  _end = data.end ();

        if (data.num_used != 0)
          {
            for (bucket_type* bucket = data.buckets; bucket != _end; ++bucket)
              {
                if (!bucket->is_empty_or_debris ()
                    || MCT_OPTIMIZATION_UNLIKELY (data.is_special (bucket)))
                  return bucket;
              }
          }

        return _end;
      }

      template <typename Data>
      typename Data::bucket_type*
      next (const Data& data)
      {
        typedef  typename Data::bucket_type  bucket_type;

        bucket_type*  bucket = static_cast <bucket_type*> (this);
        for (bucket_type* const limit = data.end (); MCT_OPTIMIZATION_LIKELY (++bucket != limit);)
          {
            if (!bucket->is_empty_or_debris ()
                || MCT_OPTIMIZATION_UNLIKELY (data.is_special (bucket)))
              break;
          }

        return bucket;
      }

      template <typename Data>
      typename Data::bucket_type*
      previous (const Data& data)
      {
        typedef  typename Data::bucket_type  bucket_type;

        for (bucket_type* bucket = static_cast <bucket_type*> (this) - 1; ; --bucket)
          {
            if (!bucket->is_empty_or_debris ()
                || MCT_OPTIMIZATION_UNLIKELY (data.is_special (bucket)))
              return bucket;
          }
      }


      void
      initialize_as_sentinel ()
      { }
    };


    template <typename Traits>
    class bucket_base <Traits, true>
    {
    public:

      typedef  typename Traits::value_type             value_type;
      typedef  typename Traits::assignable_value_type  assignable_value_type;
      typedef  typename Traits::key_type               key_type;

      static  const bool  KEEPS_HASHES             = true;
      static  const bool  SPECIAL_EMPTY_AND_DEBRIS = false;
      static  const bool  USES_SENTINEL            = true;
      static  const bool  ITERATORS_NEED_DATA      = false;


    private:

      std::size_t  _hash;
      value_type   _value;


    public:

      static  const key_type&
      extract_key (const value_type& value)
      {  return Traits::extract_key (value);  }

      const key_type&
      key () const
      {  return Traits::extract_key (_value);  }


      value_type&
      value ()
      {  return _value;  }

      const value_type&
      value () const
      {  return _value;  }

      template <typename Allocator>
      void
      construct_value (Allocator& allocator, const value_type& initializer)
      {  allocator.construct (&_value, initializer);  }

#   if MCT_CXX0X_SUPPORTED

      template <typename Allocator>
      void
      construct_value (Allocator& allocator, value_type&& initializer)
      {  allocator.construct (&_value, std::forward <value_type> (initializer));  }

#   endif


      template <typename Hash>
      std::size_t
      key_hash (const Hash& hasher) const
      {  return _hash;  }

      static  std::size_t
      postprocess_hash (std::size_t hash)
      {
        return MCT_OPTIMIZATION_LIKELY (hash >= 2) ? hash : hash + 2;
      }


      // 'hash' is guaranteed to be different from empty/debris mark, so just comparing
      // with that is both 'used' and 'matches' test at the same time.
      bool
      used_and_matches (std::size_t hash) const
      {  return this->_hash == hash;  }

      bool
      is_empty () const
      {  return _hash == 0;  }

      bool
      is_debris () const
      {  return _hash == 1;  }

      bool
      is_empty_or_debris () const
      {  return _hash <= 1;  }


      void
      store_hash (std::size_t hash)
      {  this->_hash = hash;  }

      void
      copy_hash (const bucket_base& that)
      {  this->_hash = that._hash;  }


      void
      clear ()
      {  _hash = 0;  }

      void
      destroy ()
      {  _hash = 1;  }


      // Don't do anything because 'hash' field is not changed anyway.
      void
      revert_to_empty ()
      { }

      void
      revert_to_debris ()
      { }


      template <typename Data>
      static  typename Data::bucket_type*
      first (const Data& data)
      {
        typedef  typename Data::bucket_type  bucket_type;

        bucket_type* const  _end = data.end ();

        if (data.num_used != 0)
          {
            for (bucket_type* bucket = data.buckets; bucket != _end; ++bucket)
              {
                if (!bucket->is_empty_or_debris ())
                  return bucket;
              }
          }

        return _end;
      }

      bucket_base*
      next ()
      {
        for (bucket_base* bucket = this + 1; ; ++bucket)
          {
            // Because of the sentinel, this will stop for valid iterators.
            if (!bucket->is_empty_or_debris ())
              return bucket;
          }
      }

      bucket_base*
      previous ()
      {
        for (bucket_base* bucket = this - 1; ; --bucket)
          {
            // Because of the sentinel, this will stop for valid iterators (except for
            // table begin, but decrementing that is undefined behavior anyway).
            if (!bucket->is_empty_or_debris ())
              return bucket;
          }
      }

      template <typename Data>
      typename Data::bucket_type*
      next (const Data& data)
      {
        return static_cast <typename Data::bucket_type*> (next ());
      }

      template <typename Data>
      typename Data::bucket_type*
      previous (const Data& data)
      {
        return static_cast <typename Data::bucket_type*> (previous ());
      }


      void
      initialize_as_sentinel ()
      {
        // Any value that makes it appear not clear and not destroyed will do.
        _hash = 2;
      }
    };


    template <typename Traits, bool keep_hashes>
    class plain_bucket_base : public bucket_base <Traits, keep_hashes>
    {
    public:

      static  const bool  USES_LINKS = false;


      void
      link (plain_bucket_base* before)
      { }

      void
      unlink ()
      { }

#   if MCT_DEBUGGING_MEMBERS
      static  void
      validate_links (const plain_bucket_base* sentinel, std::size_t num_used)
      { }
#   endif
    };


    template <typename Traits, bool keep_hashes>
    class linked_bucket_base : public bucket_base <Traits, keep_hashes>
    {
      typedef  bucket_base <Traits, keep_hashes>  base_type;


    public:

      static  const bool  USES_LINKS               = true;
      // While this is theoretically unsound, I'm pretty sure in real world 0 and 1 cannot
      // point to a real bucket structure.
      static  const bool  SPECIAL_EMPTY_AND_DEBRIS = false;
      static  const bool  USES_SENTINEL            = true;
      static  const bool  ITERATORS_NEED_DATA      = false;


    private:

#   if MCT_SIZEOF_POINTER == MCT_SIZEOF_INT
      typedef  unsigned int        next_storage_type;
#   elif MCT_SIZEOF_POINTER == MCT_SIZEOF_LONG
      typedef  unsigned long       next_storage_type;
#   elif MCT_HAVE_LONG_LONG && MCT_SIZEOF_POINTER == MCT_SIZEOF_LONG_LONG
      typedef  unsigned long long  next_storage_type;
#   else
#     error cannot pick an integral type with the same size as pointer
#   endif

      union
      {
        linked_bucket_base*  _next;
        next_storage_type    _next_value;
      };

      linked_bucket_base*    _previous;


    public:

      static  std::size_t
      postprocess_hash (std::size_t hash)
      {  return hash;  }


      bool
      used_and_matches (std::size_t hash) const
      {
        return (!is_empty_or_debris ()
                && (!base_type::KEEPS_HASHES || base_type::used_and_matches (hash)));
      }

      bool
      is_empty () const
      {  return _next_value == 0;  }

      bool
      is_debris () const
      {  return _next_value == 1;  }

      bool
      is_empty_or_debris () const
      {  return _next_value <= 1;  }


      void
      link (linked_bucket_base* before)
      {
        _next            = before;
        _previous        = before->_previous;
        _previous->_next = this;
        _next->_previous = this;
      }

      void
      unlink ()
      {
        _previous->_next = _next;
        _next->_previous = _previous;
      }

#   if MCT_DEBUGGING_MEMBERS

      static  void
      validate_links (const linked_bucket_base* sentinel, std::size_t num_used)
      {
        if (num_used == 0)
          {
            if (sentinel->_next != sentinel || sentinel->_previous != sentinel)
              throw std::logic_error ("wrong link sentinel in an empty table");
            return;
          }

        std::size_t          loop_size = 0;
        linked_bucket_base*  bucket    = sentinel->_next;

        for (; loop_size != num_used; ++loop_size, bucket = bucket->_next)
          {
            if (bucket == sentinel)
              throw std::logic_error ("link loop doesn't contain all the values in the table");

            if (bucket->_next->_previous != bucket || bucket->_previous->_next != bucket)
              throw std::logic_error ("broken link structure");
          }

        if (bucket != sentinel)
          throw std::logic_error ("link loop size is larger than number of values in the table");
      }

#   endif  // MCT_DEBUGGING_MEMBERS

      void
      reverse_links ()
      {  std::swap (_next, _previous);  }


      void
      clear ()
      {  _next_value = 0;  }

      void
      destroy ()
      {  _next_value = 1;  }


      // Don't do anything because '_next_value' field is not changed.
      void
      revert_to_empty ()
      { }

      void
      revert_to_debris ()
      { }


      template <typename Data>
      static  typename Data::bucket_type*
      first (const Data& data)
      {
        return static_cast <typename Data::bucket_type*> (data.end ()->_next);
      }

      linked_bucket_base*
      next ()
      {
        return _next;
      }

      linked_bucket_base*
      previous ()
      {
        return _previous;
      }


      template <typename Data>
      typename Data::bucket_type*
      next (const Data& data)
      {
        return static_cast <typename Data::bucket_type*> (_next);
      }

      template <typename Data>
      typename Data::bucket_type*
      previous (const Data& data)
      {
        return static_cast <typename Data::bucket_type*> (_previous);
      }


      void
      initialize_as_sentinel ()
      {
        _next     = this;
        _previous = this;
      }
    };


    // Note that 'buckets' and 'num_buckets' are not initialized.
    template <typename Bucket, bool with_special_buckets = Bucket::SPECIAL_EMPTY_AND_DEBRIS>
    struct closed_hash_table_data_base;


    template <typename Bucket>
    struct closed_hash_table_data_base <Bucket, false>
    {
      typedef  Bucket  bucket_type;

      bucket_type*  buckets;
      std::size_t   num_buckets;
      std::size_t   num_used;
      std::size_t   num_occupied;


      closed_hash_table_data_base ()
        : num_used     (0),
          num_occupied (0)
      { }


      bool
      is_special (const bucket_type* bucket) const
      {  return false;  }

      void
      note_new (bucket_type* bucket)
      { }

      void
      note_being_destroyed (bucket_type* bucket)
      { }


      void
      clear ()
      {
        num_used     = 0;
        num_occupied = 0;
      }


      void
      swap (closed_hash_table_data_base& that)
      {
        std::swap (buckets,      that.buckets);
        std::swap (num_buckets,  that.num_buckets);
        std::swap (num_used,     that.num_used);
        std::swap (num_occupied, that.num_occupied);
      }
    };


    template <typename Bucket>
    struct closed_hash_table_data_base <Bucket, true>
      : public closed_hash_table_data_base <Bucket, false>
    {
      typedef  closed_hash_table_data_base <Bucket, false>  base_type;

      typedef  Bucket                                       bucket_type;


      bucket_type*  special_buckets[2];


      closed_hash_table_data_base ()
      {
        special_buckets[0] = 0;
        special_buckets[1] = 0;
      }


      bool
      is_special (const bucket_type* bucket) const
      {
        return MCT_OPTIMIZATION_UNLIKELY (bucket    == special_buckets[0]
                                          || bucket == special_buckets[1]);
      }


      void
      note_new (bucket_type* bucket)
      {
        if (MCT_OPTIMIZATION_UNLIKELY (bucket->is_empty_or_debris ()))
          special_buckets[bucket->is_empty () ? 0 : 1] = bucket;
      }

      void
      note_being_destroyed (bucket_type* bucket)
      {
        if (MCT_OPTIMIZATION_UNLIKELY (bucket->is_empty_or_debris ()))
          special_buckets[bucket->is_empty () ? 0 : 1] = 0;
      }


      void
      clear ()
      {
        base_type::clear ();
        special_buckets[0] = 0;
        special_buckets[1] = 0;
      }


      void
      swap (closed_hash_table_data_base& that)
      {
        base_type::swap (that);

        // At least VC 2005 cannot swap arrays, bug #549067.
        std::swap (special_buckets[0], that.special_buckets[0]);
        std::swap (special_buckets[1], that.special_buckets[1]);
      }
    };


    template <typename Bucket>
    struct closed_hash_table_data : public closed_hash_table_data_base <Bucket>
    {
      typedef  closed_hash_table_data_base <Bucket>  base_type;

      typedef  Bucket                                bucket_type;


      std::size_t
      start_probing (std::size_t hash) const
      {
        // We require that number of buckets is a power of 2.  To compensate, we involve a
        // prime number here; this usually improves speed with particularly bad hash
        // function choices.
#     if MCT_SIZEOF_STD_SIZE_T == 4
        return (hash * 0x3874e607) & (base_type::num_buckets - 1);
#     else
        return (hash * 0x1a4548ec34f6e06f) & (base_type::num_buckets - 1);
#     endif
      }

      std::size_t
      continue_probing (std::size_t looking_at, std::size_t iteration) const
      {
        // Quadratic probing.
        return (looking_at + iteration) & (base_type::num_buckets - 1);
      }


      bucket_type*
      begin () const
      {
        return bucket_type::first (*this);
      }

      bucket_type*
      end () const
      {
        return base_type::buckets + base_type::num_buckets;
      }


#   if MCT_DEBUGGING_MEMBERS

      bool
      valid_bucket (const bucket_type* bucket) const
      {
        return (base_type::buckets <= bucket && bucket < end ()
                // Check that it doesn't point into a middle of an actual bucket.
                && ((reinterpret_cast <const char*> (bucket)
                     - reinterpret_cast <const char*> (base_type::buckets))
                    % sizeof (bucket_type) == 0)
                && (!bucket->is_empty_or_debris () || base_type::is_special (bucket)));
      }

#   endif
    };


    // We use no-data iterators only in standard (release) mode and during testing.  If
    // preconditions checking is requested (unless when testing), _all_ iterators will
    // have '_data' member, because it is required to robustly validate iterators,
    // especially before decrementing them.

# if !MCT_CHECK_PRECONDITIONS || MCT_USE_EFFICIENT_ITERATORS


    // Note that types that use special bucket pointer _need_ to have a pointer to data in
    // the iterator, because otherwise it is impossible to tell 'special' buckets from
    // simply empty/debris buckets.
    template <typename Bucket, bool keep_data_pointer = Bucket::ITERATORS_NEED_DATA>
    class hash_table_iterator_base;

    template <typename Bucket>
    class hash_table_iterator_base <Bucket, false>
    {
      typedef  Bucket  bucket_type;

    protected:

      bucket_type*  _bucket;

    public:

      typedef  std::bidirectional_iterator_tag         iterator_category;
      typedef  std::ptrdiff_t                          difference_type;
      typedef  std::size_t                             size_type;

      // Not documented; used only in MCT tests.
      static  const bool  ROBUST_PRECONDITION_CHECKING = false;


      hash_table_iterator_base ()
      { }

      hash_table_iterator_base (const closed_hash_table_data <bucket_type>& data,
                                bucket_type* bucket)
        : _bucket (bucket)
      { }


    protected:

      void
      move_forward ()
      {
        MCT_PRECONDITION (valid (), "invalid iterator");
        _bucket = static_cast <bucket_type*> (_bucket->next ());
      }

      void
      move_backward ()
      {
        // Cannot check anything without knowing the table the iterator belongs to.
        _bucket = static_cast <bucket_type*> (_bucket->previous ());
      }

      bool
      valid () const
      {
        // Required, but not enough: will not catch cases where '_bucket' points outside
        // its hashtable (most importantly, past-the-end iterator).
        return !_bucket->is_empty_or_debris ();
      }


    public:

      // Supposedly not for outsiders to use.

      bucket_type*
      bucket () const
      {  return _bucket;  }

      bool
      is_for (const closed_hash_table_data <bucket_type>& data) const
      {  return true;  }
    };


# else  // MCT_CHECK_PRECONDITIONS && !MCT_USE_EFFICIENT_ITERATORS

    // See comment earlier for explanation.
    template <typename Bucket, bool keep_data_pointer = true>
    class hash_table_iterator_base;

# endif


    // Iterators with '_data' members are defined both when precondition checking is used
    // and when not.
    template <typename Bucket>
    class hash_table_iterator_base <Bucket, true>
    {
      typedef  Bucket  bucket_type;

    protected:

      const closed_hash_table_data <bucket_type>*  _data;
      bucket_type*                                 _bucket;

    public:

      typedef  std::bidirectional_iterator_tag         iterator_category;
      typedef  std::ptrdiff_t                          difference_type;
      typedef  std::size_t                             size_type;

      // Not documented; used only in MCT tests.
      static  const bool  ROBUST_PRECONDITION_CHECKING = true;


      hash_table_iterator_base ()
      { }

      hash_table_iterator_base (const closed_hash_table_data <bucket_type>& data,
                                bucket_type* bucket)
        : _data   (&data),
          _bucket (bucket)
      { }


    protected:

      void
      move_forward ()
      {
        MCT_PRECONDITION (valid (), "invalid iterator");
        _bucket = _bucket->next (*_data);
      }

      void
      move_backward ()
      {
        MCT_PRECONDITION (_bucket == _data->end () || (_bucket != _data->begin () && valid ()),
                          "iterator not valid for decrementing");
        _bucket = _bucket->previous (*_data);
      }

      bool
      valid () const
      {  return _data->valid_bucket (_bucket);  }


    public:

      // Supposedly not for outsiders to use.

      bucket_type*
      bucket () const
      {  return _bucket;  }

      bool
      is_for (const closed_hash_table_data <bucket_type>& data) const
      {  return &data == _data;  }
    };


    template <typename Bucket>
    class hash_table_const_iterator : public hash_table_iterator_base <Bucket>
    {
      typedef  Bucket                                  bucket_type;
      typedef  hash_table_iterator_base <bucket_type>  base_type;

    public:

      typedef  const typename bucket_type::value_type  value_type;
      typedef  value_type*                             pointer;
      typedef  value_type&                             reference;


      hash_table_const_iterator ()
      { }

      hash_table_const_iterator (const closed_hash_table_data <bucket_type>& data,
                                 bucket_type* bucket)
        : base_type (data, bucket)
      { }


      value_type&
      operator* () const
      {
        MCT_PRECONDITION (base_type::valid (), "invalid iterator");
        return base_type::_bucket->value ();
      }

      value_type*
      operator-> () const
      {
        MCT_PRECONDITION (base_type::valid (), "invalid iterator");
        return &base_type::_bucket->value ();
      }


      hash_table_const_iterator&
      operator++ ()
      {
        base_type::move_forward ();
        return *this;
      }

      hash_table_const_iterator
      operator++ (int)
      {
        const hash_table_const_iterator  copy (*this);

        base_type::move_forward ();
        return copy;
      }

      hash_table_const_iterator&
      operator-- ()
      {
        base_type::move_backward ();
        return *this;
      }

      hash_table_const_iterator
      operator-- (int)
      {
        const hash_table_const_iterator  copy (*this);

        base_type::move_backward ();
        return copy;
      }
    };


    template <typename Bucket>
    class hash_table_iterator : public hash_table_const_iterator <Bucket>
    {
      typedef  Bucket                                   bucket_type;
      typedef  hash_table_const_iterator <bucket_type>  base_type;

    public:

      typedef  typename bucket_type::assignable_value_type  value_type;
      typedef  value_type*                                  pointer;
      typedef  value_type&                                  reference;


      hash_table_iterator ()
      { }

      hash_table_iterator (const closed_hash_table_data <bucket_type>& data, bucket_type* bucket)
        : base_type (data, bucket)
      { }


      value_type&
      operator* () const
      {
        MCT_PRECONDITION (base_type::valid (), "invalid iterator");
        return base_type::_bucket->value ();
      }

      value_type*
      operator-> () const
      {
        MCT_PRECONDITION (base_type::valid (), "invalid iterator");
        return &base_type::_bucket->value ();
      }


      hash_table_iterator&
      operator++ ()
      {
        return static_cast <hash_table_iterator&> (base_type::operator++ ());
      }

      hash_table_iterator
      operator++ (int)
      {
        return static_cast <hash_table_iterator> (base_type::operator++ ());
      }

      hash_table_iterator&
      operator-- ()
      {
        return static_cast <hash_table_iterator&> (base_type::operator-- ());
      }

      hash_table_iterator
      operator-- (int)
      {
        return static_cast <hash_table_iterator> (base_type::operator-- ());
      }
    };


    template <typename Bucket>
    inline  bool
    operator== (const hash_table_const_iterator <Bucket>& iterator1,
                const hash_table_const_iterator <Bucket>& iterator2)
    {  return iterator1.bucket () == iterator2.bucket ();  }

    template <typename Bucket>
    inline  bool
    operator!= (const hash_table_const_iterator <Bucket>& iterator1,
                const hash_table_const_iterator <Bucket>& iterator2)
    {  return iterator1.bucket () != iterator2.bucket ();  }

    template <typename Bucket>
    inline  bool
    operator< (const hash_table_const_iterator <Bucket>& iterator1,
               const hash_table_const_iterator <Bucket>& iterator2)
    {  return iterator1.bucket () < iterator2.bucket ();  }

    template <typename Bucket>
    inline  bool
    operator<= (const hash_table_const_iterator <Bucket>& iterator1,
                const hash_table_const_iterator <Bucket>& iterator2)
    {  return iterator1.bucket () <= iterator2.bucket ();  }

    template <typename Bucket>
    inline  bool
    operator> (const hash_table_const_iterator <Bucket>& iterator1,
               const hash_table_const_iterator <Bucket>& iterator2)
    {  return iterator1.bucket () > iterator2.bucket ();  }

    template <typename Bucket>
    inline  bool
    operator>= (const hash_table_const_iterator <Bucket>& iterator1,
                const hash_table_const_iterator <Bucket>& iterator2)
    {  return iterator1.bucket () >= iterator2.bucket ();  }


    const std::size_t  DEFAULT_NUM_BUCKETS     = 16;
    const std::size_t  MIN_NUM_BUCKETS         = 4;
    const float        DEFAULT_MAX_LOAD_FACTOR = 0.6f;
    const float        MIN_MAX_LOAD_FACTOR     = 0.01f;
    const float        MAX_MAX_LOAD_FACTOR     = 0.99f;


    template <typename Buckets, bool keep_hashes>
    struct bucket_type_picker
    {
      typedef  typename Buckets::without_hash  type;
    };

    template <typename Buckets>
    struct bucket_type_picker <Buckets, true>
    {
      typedef  typename Buckets::with_hash  type;
    };


    template <typename Traits, bool keep_hashes>
    class bucket_base;


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    class closed_hash_table
    {
      template <typename _Buckets, typename _Hash, typename _Equal, typename _Allocator,
                bool _keep_hashes>
      friend  bool
      operator== (const closed_hash_table <_Buckets, _Hash, _Equal, _Allocator,
                                           _keep_hashes>& table1,
                  const closed_hash_table <_Buckets, _Hash, _Equal, _Allocator,
                                           _keep_hashes>& table2);


    protected:

      typedef  typename bucket_type_picker <Buckets, keep_hashes>::type  bucket_type;
      typedef  closed_hash_table_data <bucket_type>                      data_type;

    public:

      typedef  typename bucket_type::key_type                            key_type;
      typedef  typename bucket_type::value_type                          value_type;
      typedef  Hash                                                      hasher;
      typedef  Equal                                                     key_equal;
      typedef  typename Allocator::template rebind <value_type>::other   allocator_type;

      typedef  hash_table_iterator <bucket_type>                         iterator;
      typedef  hash_table_const_iterator <bucket_type>                   const_iterator;

      typedef  typename allocator_type::pointer                          pointer;
      typedef  typename allocator_type::const_pointer                    const_pointer;
      typedef  typename allocator_type::reference                        reference;
      typedef  typename allocator_type::const_reference                  const_reference;
      typedef  std::ptrdiff_t                                            difference_type;
      typedef  std::size_t                                               size_type;


# if MCT_DEBUGGING_MEMBERS

      struct statistics
      {
        double     debris_ratio;
        double     avg_present_lookup;
        size_type  max_present_lookup;
        double     avg_absent_lookup;
        size_type  max_absent_lookup;
      };

# endif


    protected:

      typedef  typename allocator_type::template rebind <bucket_type>::other
               bucket_allocator_type;


      data_type       _data;
      size_type       _max_occupied;
      hasher          _hash;
      key_equal       _equal;
      allocator_type  _allocator;
      float           _max_load_factor;


      // This structure is used to simplify exception safety code.
      struct data_grip : public data_type
      {
        closed_hash_table&  table;

        data_grip (closed_hash_table& table, size_type num_buckets)
          : table (table)
        {
          table.do_allocate_buckets (*this, num_buckets);
        }

        ~data_grip ()
        {
          table.do_deallocate_buckets (*this);
        }
      };


    public:

      closed_hash_table (size_type num_buckets, const hasher& hash, const key_equal& equal,
                         const allocator_type& allocator)
        : _hash      (hash),
          _equal     (equal),
          _allocator (allocator)
      {
        allocate_buckets (num_buckets);
        do_set_max_load_factor (DEFAULT_MAX_LOAD_FACTOR);

        MCT_VALIDATION (validate_integrity ());
      }

      closed_hash_table (const closed_hash_table& that);
      closed_hash_table (const closed_hash_table& that, const allocator_type& allocator);

#   if MCT_CXX0X_SUPPORTED

      closed_hash_table (closed_hash_table&& that)
        : _data            (that._data),
          _max_occupied    (that._max_occupied),
          _hash            (that.hash_function ()),
          _equal           (that.key_eq ()),
          _allocator       (that.get_allocator ()),
          _max_load_factor (that.max_load_factor ())
      {
        that._data.buckets = 0;
        MCT_VALIDATION (validate_integrity ());
      }

      closed_hash_table (std::initializer_list <value_type> initializer,
                         size_type num_buckets, const hasher& hash, const key_equal& equal,
                         const allocator_type& allocator)
        : _hash      (hash),
          _equal     (equal),
          _allocator (allocator)
      {
        initialize_from_range (initializer.begin (), initializer.end (), num_buckets);
      }

#   endif  // MCT_CXX0X_SUPPORTED

      template <typename InputIterator>
      closed_hash_table (InputIterator first, InputIterator last,
                         size_type num_buckets, const hasher& hash, const key_equal& equal,
                         const allocator_type& allocator)
        : _hash      (hash),
          _equal     (equal),
          _allocator (allocator)
      {
        initialize_from_range (first, last, num_buckets);
      }

    private:

      template <typename InputIterator>
      void  initialize_from_range (InputIterator first, InputIterator last, size_type num_buckets);

      void  initialize_as_copy (const closed_hash_table& that);

    public:

      ~closed_hash_table ();


      closed_hash_table&  operator= (const closed_hash_table& that);

#   if MCT_CXX0X_SUPPORTED

      closed_hash_table&  operator= (closed_hash_table&& that);

      closed_hash_table&
      operator= (std::initializer_list <value_type> initializer)
      {
        closed_hash_table  temp (initializer, 0, _hash, _equal, _allocator);
        swap (temp);
        return *this;
      }

#   endif


      iterator
      begin ()
      {
        return make_iterator (_data.begin ());
      }

      const_iterator
      begin () const
      {
        return make_const_iterator (_data.begin ());
      }

      const_iterator
      cbegin () const
      {
        return make_const_iterator (_data.begin ());
      }

      iterator
      end ()
      {
        return make_iterator (_data.end ());
      }

      const_iterator
      end () const
      {
        return make_const_iterator (_data.end ());
      }

      const_iterator
      cend () const
      {
        return make_const_iterator (_data.end ());
      }


      bool
      empty () const
      {  return _data.num_used == 0;  }

      size_type
      size () const
      {  return _data.num_used;  }

      size_type
      max_size () const
      {
        return static_cast <size_type> (std::floor (max_bucket_count () * MAX_MAX_LOAD_FACTOR));
      }


      float
      load_factor () const
      {
        return static_cast <float> (_data.num_used) / _data.num_buckets;
      }

      float
      max_load_factor () const
      {  return _max_load_factor;  }

      void
      max_load_factor (float max_load_factor)
      {
        do_set_max_load_factor (max_load_factor);
        if (_data.num_occupied > _max_occupied)
          clear_debris_or_grow ();
      }


      size_type
      bucket_count () const
      {  return _data.num_buckets;  }

      size_type
      max_bucket_count () const
      {
        return round_down_to_power_of_2 (bucket_allocator_type (_allocator).max_size ()
                                         - bucket_type::USES_SENTINEL);
      }


      hasher
      hash_function () const
      {  return _hash;  }

      key_equal
      key_eq () const
      {  return _equal;  }

      allocator_type
      get_allocator () const
      {  return _allocator;  }


      iterator
      find (const key_type& key)
      {
        return make_iterator (lookup (key));
      }

      const_iterator
      find (const key_type& key) const
      {
        return make_const_iterator (lookup (key));
      }

      size_type
      count (const key_type& key) const
      {
        return lookup (key) != _data.end () ? 1 : 0;
      }

      std::pair <iterator, iterator>
      equal_range (const key_type& key)
      {
        const iterator  first (make_iterator (lookup (key)));
        iterator        last  (first);

        if (first.bucket () != _data.end ())
          ++last;

        return std::make_pair (first, last);
      }

      std::pair <const_iterator, const_iterator>
      equal_range (const key_type& key) const
      {
        const const_iterator  first (make_const_iterator (lookup (key)));
        const_iterator        last  (first);

        if (first.bucket () != _data.end ())
          ++last;

        return std::make_pair (first, last);
      }


      std::pair <iterator, bool>
      insert (const value_type& value)
      {
        const std::pair <bucket_type*, bool>  result
          (lookup_or_insert <const value_type&> (value));
        return std::make_pair (make_iterator (result.first), result.second);
      }

      // Note that we currently just ignore 'hint'.
      iterator
      insert (const_iterator hint, const value_type& value)
      {
        MCT_PRECONDITION (hint == end () || valid_iterator (hint), "invalid hint");
        return make_iterator (lookup_or_insert <const value_type&> (value).first);
      }

      template <typename InputIterator>
      void
      insert (InputIterator first, InputIterator last)
      {
        rehash_for_insertion (forward_distance (first, last));
        for (; first != last; ++first)
          lookup_or_insert <const value_type&> (*first);
      }

#   if MCT_CXX0X_SUPPORTED

      std::pair <iterator, bool>
      insert (value_type&& value)
      {
        const std::pair <bucket_type*, bool>  result
          (lookup_or_insert <value_type&&> (std::move (value)));
        return std::make_pair (make_iterator (result.first), result.second);
      }

      iterator
      insert (const_iterator hint, value_type&& value)
      {
        MCT_PRECONDITION (hint == end () || valid_iterator (hint), "invalid hint");
        return make_iterator (lookup_or_insert <value_type&&> (std::move (value)).first);
      }

      void
      insert (std::initializer_list <value_type> initializer)
      {
        insert (initializer.begin (), initializer.end ());
      }


      template <typename... Args>
      std::pair <iterator, bool>
      emplace (Args&&... args)
      {
        return insert (value_type (std::forward <Args> (args)...));
      }

      template <typename... Args>
      std::pair <iterator, bool>
      emplace_hint (const_iterator hint, Args&&... args)
      {
        MCT_PRECONDITION (hint == end () || valid_iterator (hint), "invalid hint");
        return insert (hint, value_type (std::forward <Args> (args)...));
      }

#   endif  // MCT_CXX0X_SUPPORTED


      size_type
      erase (const key_type& key)
      {
        return do_erase (key) ? 1 : 0;
      }

      iterator
      erase (const_iterator position)
      {
        MCT_PRECONDITION (valid_iterator (position), "invalid position");
        bucket_type* const  next = position.bucket ()->next (_data);
        do_erase (position.bucket ());
        return make_iterator (next);
      }

      iterator
      erase (const_iterator first, const_iterator last)
      {
        MCT_PRECONDITION (valid_iterator_range (first, last), "invalid range");

        bucket_type*  bucket = first.bucket ();
        for (bucket_type* const limit = last.bucket (); bucket != limit;)
          {
            bucket_type* const  next = bucket->next (_data);
            do_erase (bucket);
            bucket = next;
          }

        return make_iterator (bucket);
      }


      void  clear ();
      void  swap (closed_hash_table& that);

      void
      rehash (size_type num_buckets)
      {  do_rehash_hint (num_buckets);  }

      void
      reserve (size_type num_elements)
      {  do_rehash_hint (std::ceil (num_elements / max_load_factor ()));  }


#   if MCT_DEBUGGING_MEMBERS

      bool
      valid_iterator (const_iterator position) const
      {
        // In some unlikely cases certain iterators (e.g. from another table after several
        // rehashes) could point to a valid bucket, but still be invalid.
        return position.is_for (_data) && _data.valid_bucket (position.bucket ());
      }

      bool
      valid_iterator_range (const_iterator first, const_iterator last) const
      {
        const const_iterator  _end = end ();
        if (first != _end && !valid_iterator (first))
          return false;

        if (last != _end && !valid_iterator (last))
          return false;

        for (; first != _end; ++first)
          {
            if (first == last)
              return true;
          }

        return true;
      }

      size_type
      used_memory () const
      {
        return (sizeof *this
                + (_data.num_buckets + bucket_type::USES_SENTINEL) * sizeof (bucket_type));
      }

      void        validate_integrity () const;
      statistics  collect_statistics () const;

#   endif  // MCT_DEBUGGING_MEMBERS


    protected:

      iterator
      make_iterator (bucket_type* bucket) const
      {  return iterator (_data, bucket);  }

      const_iterator
      make_const_iterator (bucket_type* bucket) const
      {  return const_iterator (_data, bucket);  }


      bucket_type*  lookup (const key_type& key)  const;

      // Templated for convenience of map's operator[].
      template <typename type>
      std::pair <bucket_type*, bool>  lookup_or_insert (type data, bucket_type* before = 0);


      void  do_set_max_load_factor (float max_load_factor);


      bool
      do_erase (const key_type& key)
      {
        bucket_type* const  bucket = lookup (key);

        if (bucket != _data.end ())
          {
            do_erase (bucket);
            return true;
          }
        else
          return false;
      }

      void
      do_erase (bucket_type* bucket)
      {
        _data.note_being_destroyed (bucket);
        --_data.num_used;
        _allocator.destroy (&bucket->value ());
        bucket->unlink ();
        bucket->destroy ();

        MCT_VALIDATION (validate_integrity ());
      }


      bucket_type*
      rehash_for_insertion (size_type num_elements, bucket_type* watch_for = 0)
      {
        if (num_elements && _data.num_used + num_elements > _max_occupied)
          {
            return do_rehash_hint (static_cast <size_type> ((_data.num_used + num_elements)
                                                            / max_load_factor ()),
                                   watch_for);
          }
        else
          return watch_for;
      }

      bucket_type*  do_rehash_hint  (size_type num_buckets, bucket_type* watch_for = 0);
      bucket_type*  do_rehash_final (size_type num_buckets, bucket_type* watch_for = 0);
      bucket_type*  clear_debris_or_grow (bucket_type* watch_for = 0);


      size_type
      compute_max_occupied () const
      {
        if (_data.num_buckets < max_bucket_count ())
          {
            return std::min (_data.num_buckets - 1,
                             static_cast <size_type> (_data.num_buckets * _max_load_factor));
          }
        else
          return max_size ();
      }


    private:

      void
      allocate_buckets (size_type num_buckets)
      {
        if (num_buckets > MIN_NUM_BUCKETS)
          num_buckets = round_up_to_power_of_2 (num_buckets);
        else
          num_buckets = (num_buckets ? MIN_NUM_BUCKETS : DEFAULT_NUM_BUCKETS);

        do_allocate_buckets (_data, num_buckets);
      }

      void
      do_allocate_buckets (data_type& data, size_type num_buckets)
      {
        if (num_buckets > max_bucket_count ())
          throw std::length_error ("closed_hash_table::do_allocate_buckets");

        const size_type  real_num_buckets = (num_buckets + bucket_type::USES_SENTINEL);

        data.num_buckets = num_buckets;
        data.buckets     = bucket_allocator_type (_allocator).allocate (real_num_buckets);

        for (bucket_type* bucket = data.buckets, * const limit = data.end ();
             bucket != limit; ++bucket)
          bucket->clear ();

        if (bucket_type::USES_SENTINEL)
          data.end ()->initialize_as_sentinel ();
      }

      void
      do_deallocate_buckets (const data_type& data)
      {
        // Note: we mustn't depend on exact value of 'num_occupied', because that might be
        // incorrect if we are cleaning up after an exception.  However, it will not be
        // zero if the table is or can be non-empty.
        if (!has_trivial_destructor <value_type>::value && data.num_occupied)
          {
            for (bucket_type* bucket = data.buckets, * const limit = data.end ();
                 bucket != limit; ++bucket)
              {
                if (!bucket->is_empty_or_debris () || data.is_special (bucket))
                  _allocator.destroy (&bucket->value ());
              }
          }

        const size_type  real_num_buckets = (data.num_buckets + bucket_type::USES_SENTINEL);
        bucket_allocator_type (_allocator).deallocate (data.buckets, real_num_buckets);
      }


      // This is really internal.  Destination bucket array must be completely clean and
      // have enough space to fit the used elements in 'from'.  'from' must have been
      // built using the same equality and hash functions as this table.  Returned value
      // is the bucket to where 'watch_for' is copied.
      bucket_type*
      copy_buckets (const data_type& from, data_type& to, bucket_type* watch_for = 0);
    };


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    class linked_hash_table
      : public closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>
    {
      typedef  closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>  base_type;

    protected:

      typedef  typename base_type::bucket_type                                   bucket_type;

    public:

      typedef  typename base_type::value_type          value_type;
      typedef  typename base_type::hasher              hasher;
      typedef  typename base_type::key_equal           key_equal;
      typedef  typename base_type::allocator_type      allocator_type;
      typedef  typename base_type::iterator            iterator;
      typedef  typename base_type::const_iterator      const_iterator;
      typedef  std::reverse_iterator <iterator>        reverse_iterator;
      typedef  std::reverse_iterator <const_iterator>  const_reverse_iterator;
      typedef  typename base_type::size_type           size_type;


      linked_hash_table (size_type num_buckets, const hasher& hash, const key_equal& equal,
                         const allocator_type& allocator)
        : base_type (num_buckets, hash, equal, allocator)
      { }

      linked_hash_table (const linked_hash_table& that)
        : base_type (that)
      { }

      linked_hash_table (const linked_hash_table& that, const allocator_type& allocator)
        : base_type (that, allocator)
      { }

#   if MCT_CXX0X_SUPPORTED

      linked_hash_table (linked_hash_table&& that)
        : base_type (std::move (that))
      { }

      // We don't use similar base constructor, because we need a non-default second
      // argument to lookup_or_insert().
      linked_hash_table (std::initializer_list <value_type> initializer,
                         size_type num_buckets, const hasher& hash, const key_equal& equal,
                         const allocator_type& allocator)
        : base_type ((num_buckets
                      ? num_buckets : std::ceil (initializer.size () / DEFAULT_MAX_LOAD_FACTOR)),
                     hash, equal, allocator)
      {
        initialize_from_range (initializer.begin (), initializer.end ());
      }

#   endif  // MCT_CXX0X_SUPPORTED

      // We don't use similar base constructor, because we need a non-default second
      // argument to lookup_or_insert().
      template <typename InputIterator>
      linked_hash_table (InputIterator first, InputIterator last,
                         size_type num_buckets, const hasher& hash, const key_equal& equal,
                         const allocator_type& allocator)
        : base_type ((num_buckets
                      ? num_buckets
                      : forward_distance_scaled (first, last, DEFAULT_MAX_LOAD_FACTOR)),
                     hash, equal, allocator)
      {
        initialize_from_range (first, last);
      }


    private:

      template <typename InputIterator>
      void
      initialize_from_range (InputIterator first, InputIterator last)
      {
        for (; first != last; ++first)
          {
            base_type::template lookup_or_insert <const value_type&> (*first,
                                                                      base_type::_data.end ());
          }

        MCT_VALIDATION (base_type::validate_integrity ());
      }


    public:

#   if MCT_CXX0X_SUPPORTED

      // Since base class uses insert(), we cannot call super method here.
      linked_hash_table&
      operator= (std::initializer_list <value_type> initializer)
      {
        linked_hash_table  temp (initializer, 0,
                                 base_type::_hash, base_type::_equal, base_type::_allocator);
        swap (temp);
        return *this;
      }

#   endif


      reverse_iterator
      rbegin ()
      {
        return reverse_iterator (base_type::end ());
      }

      const_reverse_iterator
      rbegin () const
      {
        return const_reverse_iterator (base_type::end ());
      }

      const_reverse_iterator
      crbegin () const
      {
        return const_reverse_iterator (base_type::cend ());
      }

      reverse_iterator
      rend ()
      {
        return reverse_iterator (base_type::begin ());
      }

      const_reverse_iterator
      rend () const
      {
        return const_reverse_iterator (base_type::begin ());
      }

      const_reverse_iterator
      crend () const
      {
        return const_reverse_iterator (base_type::cbegin ());
      }


      value_type&
      front ()
      {
        MCT_PRECONDITION (!base_type::empty (), "table is empty");
        return base_type::_data.begin ()->value ();
      }

      const value_type&
      front () const
      {
        MCT_PRECONDITION (!base_type::empty (), "table is empty");
        return base_type::_data.begin ()->value ();
      }

      value_type&
      back ()
      {
        MCT_PRECONDITION (!base_type::empty (), "table is empty");
        return base_type::_data.end ()->previous ()->value ();
      }

      const value_type&
      back () const
      {
        MCT_PRECONDITION (!base_type::empty (), "table is empty");
        return base_type::_data.end ()->previous ()->value ();
      }


      std::pair <iterator, bool>
      push_front (const value_type& value)
      {
        const std::pair <bucket_type*, bool>  result
          (base_type::template lookup_or_insert <const value_type&> (value,
                                                                     base_type::_data.begin ()));
        return std::make_pair (make_iterator (result.first), result.second);
      }

      void
      pop_front ()
      {
        MCT_PRECONDITION (!base_type::empty (), "table is empty");
        erase (make_iterator (base_type::_data.begin ()));
      }

      std::pair <iterator, bool>
      push_back (const value_type& value)
      {
        const std::pair <bucket_type*, bool>  result
          (base_type::template lookup_or_insert <const value_type&> (value,
                                                                     base_type::_data.end ()));
        return std::make_pair (make_iterator (result.first), result.second);
      }

      void
      pop_back ()
      {
        MCT_PRECONDITION (!base_type::empty (), "table is empty");
        erase (make_iterator (base_type::_data.end ()->previous ()));
      }


#   if MCT_CXX0X_SUPPORTED

      std::pair <iterator, bool>
      push_front (value_type&& value)
      {
        const std::pair <bucket_type*, bool>  result
          (base_type::template lookup_or_insert <value_type&&> (std::move (value),
                                                                base_type::_data.begin ()));
        return std::make_pair (make_iterator (result.first), result.second);
      }

      std::pair <iterator, bool>
      push_back (value_type&& value)
      {
        const std::pair <bucket_type*, bool>  result
          (base_type::template lookup_or_insert <value_type&&> (std::move (value),
                                                                base_type::_data.end ()));
        return std::make_pair (make_iterator (result.first), result.second);
      }

#   endif  // MCT_CXX0X_SUPPORTED


      // Need to override all 'insert' overloads, else lookup fails.  Additionally, in
      // 'linked_hash_table' lookup_or_insert() actually uses 'before' argument, so 0
      // default value is not acceptable.

      std::pair <iterator, bool>
      insert (const value_type& value)
      {
        const std::pair <bucket_type*, bool>  result
          (base_type::template lookup_or_insert <const value_type&> (value,
                                                                     base_type::_data.end ()));
        return std::make_pair (make_iterator (result.first), result.second);
      }

      template <typename InputIterator>
      void
      insert (InputIterator first, InputIterator last)
      {  insert (base_type::end (), first, last);  }

      std::pair <iterator, bool>
      insert (const_iterator before, const value_type& value)
      {
        MCT_PRECONDITION (before == base_type::end () || base_type::valid_iterator (before),
                          "invalid position");
        const std::pair <bucket_type*, bool>  result
          (base_type::template lookup_or_insert <const value_type&> (value, before.bucket ()));
        return std::make_pair (make_iterator (result.first), result.second);
      }

      template <typename InputIterator>
      void  insert (const_iterator before, InputIterator first, InputIterator last);

#   if MCT_CXX0X_SUPPORTED

      std::pair <iterator, bool>
      insert (value_type&& value)
      {
        const std::pair <bucket_type*, bool>  result
          (base_type::template lookup_or_insert <value_type&&> (std::move (value),
                                                                base_type::_data.end ()));
        return std::make_pair (make_iterator (result.first), result.second);
      }

      std::pair <iterator, bool>
      insert (const_iterator before, value_type&& value)
      {
        MCT_PRECONDITION (before == base_type::end () || base_type::valid_iterator (before),
                          "invalid position");
        const std::pair <bucket_type*, bool>  result
          (base_type::template lookup_or_insert <value_type&&> (std::move (value),
                                                                before.bucket ()));
        return std::make_pair (make_iterator (result.first), result.second);
      }

      void
      insert (std::initializer_list <value_type> initializer)
      {
        insert (base_type::end (), initializer.begin (), initializer.end ());
      }

      void
      insert (const_iterator before, std::initializer_list <value_type> initializer)
      {
        insert (before, initializer.begin (), initializer.end ());
      }


      template <typename... Args>
      std::pair <iterator, bool>
      emplace (Args&&... args)
      {
        return insert (value_type (std::forward <Args> (args)...));
      }

      template <typename... Args>
      std::pair <iterator, bool>
      emplace_before (const_iterator before, Args&&... args)
      {
        return insert (before, value_type (std::forward <Args> (args)...));
      }

      template <typename... Args>
      std::pair <iterator, bool>
      emplace_hint (const_iterator hint, Args&&... args)
      {
        MCT_PRECONDITION (hint == base_type::end () || base_type::valid_iterator (hint),
                          "invalid hint");
        return insert (hint, value_type (std::forward <Args> (args)...));
      }

#   endif  // MCT_CXX0X_SUPPORTED


      void
      relink (const_iterator before, iterator element)
      {
        MCT_PRECONDITION (before == base_type::end () || base_type::valid_iterator (before),
                          "invalid position");
        MCT_PRECONDITION (base_type::valid_iterator (element), "invalid element");

        if (element.bucket () != before.bucket ())
          {
            element.bucket ()->unlink ();
            element.bucket ()->link (before.bucket ());

            MCT_VALIDATION (base_type::validate_integrity ());
          }
      }

      void  reverse ();
    };


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    closed_hash_table (const closed_hash_table& that)
      : _hash            (that.hash_function ()),
        _equal           (that.key_eq ()),
        _allocator       (that.get_allocator ()),
        _max_load_factor (that.max_load_factor ())
    {
      initialize_as_copy (that);
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    closed_hash_table (const closed_hash_table& that, const allocator_type& allocator)
      : _hash            (that.hash_function ()),
        _equal           (that.key_eq ()),
        _allocator       (allocator),
        _max_load_factor (that.max_load_factor ())
    {
      initialize_as_copy (that);
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    template <typename InputIterator>
    void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    initialize_from_range (InputIterator first, InputIterator last, size_type num_buckets)
    {
      allocate_buckets (num_buckets
                        ? num_buckets
                        : forward_distance_scaled (first, last, DEFAULT_MAX_LOAD_FACTOR));
      do_set_max_load_factor (DEFAULT_MAX_LOAD_FACTOR);

      try
        {
          for (; first != last; ++first)
            lookup_or_insert <const value_type&> (*first);
        }
      catch (...)
        {
          do_deallocate_buckets (_data);
          throw;
        }

      MCT_VALIDATION (validate_integrity ());
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    inline  void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    initialize_as_copy (const closed_hash_table& that)
    {
      allocate_buckets (static_cast <size_type> (std::ceil (that._data.num_used
                                                            / max_load_factor ())));
      _max_occupied = compute_max_occupied ();

      try
        {
          copy_buckets (that._data, _data);
        }
      catch (...)
        {
          do_deallocate_buckets (_data);
          throw;
        }

      MCT_VALIDATION (validate_integrity ());
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    ~closed_hash_table ()
    {
      // Move constructor or assignment can "steal" buckets.
      if (_data.buckets)
        do_deallocate_buckets (_data);
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>&
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    operator= (const closed_hash_table& that)
    {
      // Just an optimization; will work without this check too.
      if (&that != this)
        {
          closed_hash_table  copy (that, _allocator);
          swap (copy);
        }

      return *this;
    }

# if MCT_CXX0X_SUPPORTED

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>&
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    operator= (closed_hash_table&& that)
    {
      if (this->_allocator == that._allocator)
        {
          do_deallocate_buckets (_data);

          this->_data            = that._data;
          this->_max_occupied    = that._max_occupied;
          this->_hash            = that._hash;
          this->_equal           = that._equal;
          this->_max_load_factor = that._max_load_factor;

          that._data.buckets = 0;
        }
      else
        {
          closed_hash_table  copy (that, _allocator);
          swap (copy);
        }

      return *this;
    }

# endif  // MCT_CXX0X_SUPPORTED


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    do_set_max_load_factor (float max_load_factor)
    {
      _max_load_factor = std::max (MIN_MAX_LOAD_FACTOR,
                                   std::min (max_load_factor, MAX_MAX_LOAD_FACTOR));
      _max_occupied    = compute_max_occupied ();
    }


    // According to benchmarks, explicitly unrolling first pass gives a small speedup for
    // lookup() and lookup_or_insert().

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    typename closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::bucket_type*
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    lookup (const key_type& key) const
    {
      const size_type  hash    = bucket_type::postprocess_hash (_hash (key));
      size_type        look_at = _data.start_probing (hash);

      bucket_type* const  bucket = (_data.buckets + look_at);
      if (bucket->used_and_matches (hash) || _data.is_special (bucket))
        {
          if (_equal (bucket->key (), key))
            return bucket;
        }
      else if (bucket->is_empty ())
        return _data.end ();

      for (size_type iteration = 1; ; ++iteration)
        {
          look_at = _data.continue_probing (look_at, iteration);

          bucket_type* const  bucket = (_data.buckets + look_at);
          if (bucket->used_and_matches (hash) || _data.is_special (bucket))
            {
              if (_equal (bucket->key (), key))
                return bucket;
            }
          else if (bucket->is_empty ())
            return _data.end ();
        }
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    template <typename type>
    std::pair <typename closed_hash_table <Buckets, Hash, Equal, Allocator,
                                           keep_hashes>::bucket_type*,
               bool>
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    lookup_or_insert (type data, bucket_type* before)
    {
      const key_type&  key     = bucket_type::extract_key (data);
      const size_type  hash    = bucket_type::postprocess_hash (_hash (key));
      size_type        look_at = _data.start_probing (hash);

      bucket_type* const  bucket = (_data.buckets + look_at);
      bucket_type*        debris;

      if (bucket->used_and_matches (hash) || _data.is_special (bucket))
        {
          if (_equal (bucket->key (), key))
            return std::make_pair (bucket, false);

          debris = 0;
        }
      else if (bucket->is_empty ())
        {
          if (_data.num_occupied >= _max_occupied)
            {
              return lookup_or_insert <type> (MCT_STD_FORWARD (type, data),
                                              clear_debris_or_grow (before));
            }

          try
            {
              bucket->construct_value (_allocator, MCT_STD_FORWARD (type, data));
            }
          catch (...)
            {
              bucket->revert_to_empty ();
              throw;
            }

          bucket->store_hash (hash);
          // At least on GCC 4.4 this helps optimization.
          if (bucket_type::USES_LINKS)
            bucket->link (before);
          _data.note_new (bucket);

          ++_data.num_used;
          ++_data.num_occupied;

          MCT_VALIDATION (validate_integrity ());
          return std::make_pair (bucket, true);
        }
      else
        debris = (!bucket_type::KEEPS_HASHES || bucket->is_debris () ? bucket : 0);

      for (size_type iteration = 1; ; ++iteration)
        {
          look_at = _data.continue_probing (look_at, iteration);

          bucket_type*  bucket = (_data.buckets + look_at);

          if (bucket->used_and_matches (hash) || _data.is_special (bucket))
            {
              if (_equal (bucket->key (), key))
                return std::make_pair (bucket, false);
            }
          else if (bucket->is_empty ())
            {
              if (!debris)
                {
                  if (_data.num_occupied >= _max_occupied)
                    {
                      return lookup_or_insert <type> (MCT_STD_FORWARD (type, data),
                                                      clear_debris_or_grow (before));
                    }
                }
              else
                bucket = debris;

              try
                {
                  bucket->construct_value (_allocator, MCT_STD_FORWARD (type, data));
                }
              catch (...)
                {
                  if (bucket != debris)
                    bucket->revert_to_empty ();
                  else
                    bucket->revert_to_debris ();

                  throw;
                }

              bucket->store_hash (hash);
              if (bucket_type::USES_LINKS)
                bucket->link (before);
              _data.note_new (bucket);

              ++_data.num_used;
              if (bucket != debris)
                ++_data.num_occupied;

              MCT_VALIDATION (validate_integrity ());
              return std::make_pair (bucket, true);
            }
          else if (!debris && (!bucket_type::KEEPS_HASHES || bucket->is_debris ()))
            debris = bucket;
        }
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    clear ()
    {
      if (!_data.num_occupied)
        return;

      for (bucket_type* bucket = _data.buckets, * const limit = _data.end ();
           bucket != limit; ++bucket)
        {
          if (!has_trivial_destructor <value_type>::value)
            {
              if (!bucket->is_empty_or_debris () || _data.is_special (bucket))
                _allocator.destroy (&bucket->value ());
            }

          bucket->clear ();
        }

      _data.clear ();
      if (bucket_type::USES_SENTINEL)
        _data.end ()->initialize_as_sentinel ();

      MCT_VALIDATION (validate_integrity ());
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    swap (closed_hash_table& that)
    {
      if (this->_allocator == that._allocator)
        _data.swap (that._data);
      else
        {
          // 'new_this_data' must be allocated with 'this' allocator, but data must come
          // from 'that' and also copied by 'that' (because hash function and equality
          // predicate are not swapped yet).
          data_grip  new_this_data (*this, that ._data.num_buckets);
          data_grip  new_that_data (that,  this->_data.num_buckets);

          that .copy_buckets (that. _data, new_this_data);
          this->copy_buckets (this->_data, new_that_data);

          this->_data.swap (new_this_data);
          that ._data.swap (new_that_data);
        }

      std::swap (_max_occupied,    that._max_occupied);
      std::swap (_hash,            that._hash);
      std::swap (_equal,           that._equal);
      std::swap (_max_load_factor, that._max_load_factor);

      MCT_VALIDATION (this->validate_integrity ());
      MCT_VALIDATION (that. validate_integrity ());
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    typename closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::bucket_type*
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    do_rehash_hint (size_type num_buckets, bucket_type* watch_for)
    {
      num_buckets = std::max (num_buckets, static_cast <size_type> (_data.num_used
                                                                    / max_load_factor ()));
      num_buckets = round_up_to_power_of_2 (std::max (num_buckets, _data.num_used + 1));

      // If number of buckets isn't different still rehash to get rid of debris.
      if (num_buckets != _data.num_buckets || _data.num_occupied > _data.num_used)
        return do_rehash_final (num_buckets, watch_for);
      else
        return watch_for;
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    typename closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::bucket_type*
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    do_rehash_final (size_type num_buckets, bucket_type* watch_for)
    {
      data_grip     new_data (*this, num_buckets);
      bucket_type*  result = copy_buckets (_data, new_data, watch_for);

      _data.swap (new_data);
      _max_occupied = compute_max_occupied ();

      MCT_VALIDATION (validate_integrity ());
      return result;
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    typename closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::bucket_type*
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    clear_debris_or_grow (bucket_type* watch_for)
    {
      size_type  num_buckets = _data.num_buckets;

      // Bit shifting trick to avoid overflow if table is *huge*.
      if (_data.num_used >= (_data.num_occupied - (_data.num_occupied >> 4)))
        {
          // If almost all table is filled with real data rather than debris, we increase
          // table size instead.
          num_buckets <<= 1;
        }

      return do_rehash_final (num_buckets, watch_for);
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    typename closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::bucket_type*
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    copy_buckets (const data_type& from, data_type& to, bucket_type* watch_for)
    {
      // If copying is interrupted with exception we needn't have correct values here, but
      // we need to at least have non-zero values to bust possible is-empty optimizations.
      to.num_used     = from.num_used;
      to.num_occupied = from.num_used;

      bucket_type* const  from_end = from.end ();
      bucket_type* const  to_end   = to.end ();

      if (from.num_used)
        {
          for (bucket_type* bucket = from.begin (); bucket != from_end;
               bucket = bucket->next (from))
            {
              const size_type  hash    = bucket->key_hash (_hash);
              size_type        look_at = to.start_probing (hash);

              bucket_type*  new_bucket = (to.buckets + look_at);

              if (!new_bucket->is_empty () || to.is_special (new_bucket))
                {
                  for (size_type iteration = 1; ; ++iteration)
                    {
                      look_at = to.continue_probing (look_at, iteration);

                      new_bucket = (to.buckets + look_at);
                      if (new_bucket->is_empty () && !to.is_special (new_bucket))
                        break;
                    }
                }

              try
                {
                  new_bucket->construct_value (_allocator, bucket->value ());
                }
              catch (...)
                {
                  new_bucket->revert_to_empty ();
                  throw;
                }

              new_bucket->copy_hash (*bucket);
              new_bucket->link (to_end);
              to.note_new (new_bucket);

              if (watch_for == bucket)
                watch_for = new_bucket;
            }
        }

      if (watch_for == from_end)
        watch_for = to_end;

      return watch_for;
    }


# if MCT_DEBUGGING_MEMBERS

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    void
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    validate_integrity () const
    {
      if (_max_occupied != compute_max_occupied ())
        throw std::logic_error ("occupation threshold is incorrect");

      if (_data.num_occupied > _max_occupied)
        throw std::logic_error ("too many occupied buckets");

      size_type  num_used     = 0;
      size_type  num_occupied = 0;

      for (bucket_type* bucket = _data.buckets, * const _end = _data.end ();
           bucket != _end; ++bucket)
        {
          if (!_data.is_special (bucket))
            {
              if (bucket->is_empty ())
                continue;

              ++num_occupied;
              if (bucket->is_debris ())
                continue;
            }
          else
            ++num_occupied;

          ++num_used;
          if (lookup (bucket->key ()) != bucket)
            throw std::logic_error ("bucket lookup fails");
        }

      if (_data.num_used != num_used)
        throw std::logic_error ("wrong number of used buckets");

      if (_data.num_occupied != num_occupied)
        throw std::logic_error ("wrong number of occupied buckets");

      bucket_type::validate_links (_data.end (), num_used);
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    typename closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::statistics
    closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    collect_statistics () const
    {
      statistics  stats;

      stats.debris_ratio       = 0;
      stats.avg_present_lookup = 0;
      stats.max_present_lookup = 0;
      stats.avg_absent_lookup  = 0;
      stats.max_absent_lookup  = 0;

      if (_data.num_occupied != 0)
        {
          stats.debris_ratio = (static_cast <double> (_data.num_occupied - _data.num_used)
                                / _data.num_occupied);
        }

      for (bucket_type* bucket = _data.buckets, * const _end = _data.end ();
           bucket != _end; ++bucket)
        {
          if (!bucket->is_empty_or_debris () || _data.is_special (bucket))
            {
              const size_type  hash    = _hash (bucket->key ());
              size_type        look_at = _data.start_probing (hash);

              for (size_type iteration = 1; ; ++iteration)
                {
                  if (_data.buckets + look_at == bucket)
                    {
                      stats.avg_present_lookup += iteration;
                      stats.max_present_lookup  = std::max (stats.max_present_lookup, iteration);
                      break;
                    }

                  look_at = _data.continue_probing (look_at, iteration);
                }
            }

          size_type  look_at = (bucket - _data.buckets);

          for (size_type iteration = 1; ; ++iteration)
            {
              if ((_data.buckets + look_at)->is_empty ())
                {
                  stats.avg_absent_lookup += iteration;
                  stats.max_absent_lookup  = std::max (stats.max_absent_lookup, iteration);
                  break;
                }

              look_at = _data.continue_probing (look_at, iteration);
            }
        }

      if (_data.num_used != 0)
        stats.avg_present_lookup /= _data.num_used;

      stats.avg_absent_lookup /= _data.num_buckets;

      return stats;
    }

# endif  // MCT_DEBUGGING_MEMBERS


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    template <typename InputIterator>
    void
    linked_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    insert (const_iterator before, InputIterator first, InputIterator last)
    {
      MCT_PRECONDITION (before == base_type::end () || base_type::valid_iterator (before),
                        "invalid position");

      bucket_type*  before_bucket = base_type::rehash_for_insertion (forward_distance (first,
                                                                                       last),
                                                                     before.bucket ());
      size_type     num_inserted  = 0;

      // For linked tables it is possible to provide strong exception safety.
      // Additionally, we need cannot keep using 'before' as it may be invalidated due to
      // a rehash, which may happen at any iteration, especially for real input iterators.
      try
        {
          for (; first != last; ++first)
            {
              std::pair <bucket_type*, bool>  result
                (base_type::template lookup_or_insert <const value_type&> (*first, before_bucket));

              if (result.second)
                {
                  ++num_inserted;
                  before_bucket = static_cast <bucket_type*> (result.first->next ());
                }
            }
        }
      catch (...)
        {
          const_iterator  rollback_last  (make_const_iterator (before_bucket));
          const_iterator  rollback_first (rollback_last);

          while (num_inserted--)
            --rollback_first;

          erase (rollback_first, rollback_last);
          throw;
        }
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    void
    linked_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>::
    reverse ()
    {
      for (bucket_type* bucket = base_type::_data.end (), *limit = bucket; ;)
        {
          bucket->reverse_links ();
          bucket = static_cast <bucket_type*> (bucket->next ());
          if (bucket == limit)
            break;
        }

      MCT_VALIDATION (base_type::validate_integrity ());
    }


    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    bool
    operator== (const closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>& table1,
                const closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>& table2)
    {
      typedef  closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>  table_type;
      typedef  typename table_type::bucket_type  bucket_type;
      typedef  typename table_type::size_type    size_type;

      if (&table1 == &table2)
        return true;

      const size_type  size = table1.size ();
      if (table2.size () != size)
        return false;

      if (size > 0)
        {
          const bucket_type* const             end = table2._data.end ();
          typename table_type::const_iterator  iterator (table1.begin ());

          for (size_type k = 0; k < size; ++k, ++iterator)
            {
              const bucket_type*  match = table2.lookup (iterator.bucket ()->key ());
              if (match == end || !match->mapped_equal (*iterator.bucket ()))
                return false;
            }
        }

      return true;
    }

    template <typename Buckets, typename Hash, typename Equal, typename Allocator,
              bool keep_hashes>
    inline  bool
    operator!= (const closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>& table1,
                const closed_hash_table <Buckets, Hash, Equal, Allocator, keep_hashes>& table2)
    {
      return !operator== (table1, table2);
    }

  }

}


#undef MCT_STD_FORWARD


#endif  // Multi-inclusion guard.


// Local variables:
// mode: c++
// c-basic-offset: 2
// indent-tabs-mode: nil
// fill-column: 90
// End:
