// Copyright (c) 2010 Paul Pogonyshev.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.


#include "tests/common.hpp"

#include <map>

#include <boost/mpl/push_back.hpp>
#include <boost/mpl/vector.hpp>
#include <boost/test/unit_test.hpp>

#if HAVE_UNORDERED_MAP
#  include <unordered_map>
using std::unordered_map;

#elif HAVE_TR1_UNORDERED_MAP
#  include <tr1/unordered_map>
using std::tr1::unordered_map;

#elif HAVE_BOOST_UNORDERED_MAP_HPP
#  include <boost/unordered_map.hpp>
using boost::unordered_map;
#endif


namespace mct
{

#define TEST_MAP_OUTPUT(tested_map_type)                                                        \
  template <typename Key, typename Mapped, typename Hash, typename Equal, typename Allocator,   \
            bool keep_hashes>                                                                   \
  ostream&                                                                                      \
  operator<< (ostream& stream,                                                                  \
              const tested_map_type <Key, Mapped, Hash, Equal, Allocator, keep_hashes>& map)    \
  {                                                                                             \
    return print_map_range (stream, "{}", map.begin (), map.end ());                            \
  }

  TEST_MAP_OUTPUT (closed_hash_map)
  TEST_MAP_OUTPUT (linked_hash_map)

}


namespace std
{

  template <typename Key, typename Mapped, typename Hash, typename Equal, typename Allocator>
  ostream&
  operator<< (ostream& stream, const unordered_map <Key, Mapped, Hash, Equal, Allocator>& map)
  {
    return print_map_range (stream, "{}", map.begin (), map.end ());
  }

}


namespace
{

#define TEST_MAP_COMPARISON(tested_map_type)                                                    \
  template <typename Key, typename Mapped, typename Hash, typename Equal, typename Allocator,   \
            bool keep_hashes>                                                                   \
  bool                                                                                          \
  operator== (const tested_map_type <Key, Mapped, Hash, Equal, leak_test_allocator <Allocator>, \
                                     keep_hashes>& tested_map,                                  \
              const unordered_map <Key, Mapped, Hash, Equal, Allocator>& robust_map)            \
  {                                                                                             \
    if (tested_map.size () != robust_map.size ())                                               \
      return false;                                                                             \
    typedef  pair <const Key, Mapped>  value_type;                                              \
    foreach (const value_type& association, robust_map)                                         \
      {                                                                                         \
        typename unordered_map <Key, Mapped, Hash, Equal, Allocator>::const_iterator  match     \
          = robust_map.find (association.first);                                                \
        if (match == robust_map.end () || match->second != association.second)                  \
          return false;                                                                         \
      }                                                                                         \
    return true;                                                                                \
  }

  TEST_MAP_COMPARISON (closed_hash_map)
  TEST_MAP_COMPARISON (linked_hash_map)


  template <typename base_type>
  class test_map : public base_type
  {
  public:

    typedef  typename base_type::key_type        key_type;
    typedef  typename base_type::mapped_type     mapped_type;
    typedef  typename base_type::value_type      value_type;
    typedef  typename base_type::iterator        iterator;
    typedef  typename base_type::const_iterator  const_iterator;
    typedef  typename base_type::hasher          hasher;
    typedef  typename base_type::key_equal       key_equal;
    typedef  typename base_type::allocator_type  allocator_type;
    typedef  typename allocator_type::base_type  allocator_base;

  private:

    // Note that we don't use leak test allocator for the map we check against.
    typedef  unordered_map <key_type, mapped_type, hasher, key_equal, allocator_base>  robust_type;


    robust_type  _valid;


  public:

    explicit
    test_map (size_t                num_buckets = 0,
              const hasher&         hash        = hasher (),
              const key_equal&      equal       = key_equal (),
              const allocator_type& allocator   = allocator_type ())
      : base_type (num_buckets, hash, equal, allocator),
        _valid    (num_buckets, hash, equal, allocator)
    {
      validate ();
    }

    template <typename InputIterator>
    test_map (InputIterator begin, InputIterator end,
              size_t                num_buckets = 0,
              const hasher&         hash        = hasher (),
              const key_equal&      equal       = key_equal (),
              const allocator_type& allocator   = allocator_type ())
      : base_type (begin, end, num_buckets, hash, equal, allocator),
        _valid    (begin, end, num_buckets, hash, equal, allocator)
    {
      validate ();
    }

    ~test_map ()
    {
      validate ();
    }


    pair <iterator, bool>
    insert (const value_type& value)
    {
      const pair <iterator, bool>  result (base_type::insert (value));
      _valid.insert (value);
      validate ();
      return result;
    }

    iterator
    insert (const_iterator hint, const value_type& value)
    {
      const iterator  result (base_type::insert (hint, value));
      _valid.insert (_valid.end (), value);
      validate ();
      return result;
    }

    template <typename InputIterator>
    void
    insert (InputIterator begin, InputIterator end)
    {
      base_type::insert (begin, end);
      _valid.insert (begin, end);
      validate ();
    }

    bool
    erase (const key_type& key)
    {
      const bool  result = base_type::erase (key);
      BOOST_REQUIRE_EQUAL (result, _valid.erase (key));
      validate ();
      return result;
    }

    iterator
    erase (const_iterator position)
    {
      _valid.erase (*position);
      const iterator  result = base_type::erase (position);
      validate ();
      return result;
    }

    iterator
    erase (const_iterator first, const_iterator last)
    {
      // Order is not required to be the same; emulate.
      for (const_iterator scan = first; scan != last; ++scan)
        _valid.erase (*scan);

      const iterator  result = base_type::erase (first, last);
      validate ();
      return result;
    }


    // FIXME: This works only if tests always assign mapped_type() to the reference.
    mapped_type&
    operator[] (const key_type& key)
    {
      mapped_type&  result = base_type::operator[] (key);
      _valid[key];
      validate ();
      return result;
    }


    void
    clear ()
    {
      base_type::clear ();
      _valid.clear ();
      validate ();
    }


    void
    swap (test_map& that)
    {
      base_type::swap (that);
      _valid.swap (that._valid);
      validate ();
      that.validate ();
    }


    void
    rehash (size_t num_buckets)
    {
      base_type::rehash (num_buckets);
      validate ();
    }


    void
    validate () const
    {
      base_type::validate_integrity ();
      BOOST_REQUIRE_EQUAL (*this, _valid);
    }
  };


  template <typename base_type>
  inline  void
  swap (test_map <base_type>& map1, test_map <base_type>& map2)
  {
    map1.swap (map2);
  }


  struct Plain
  {
    template <typename Key, typename Mapped, typename Hash, typename Equal, bool keep_hashes>
    struct resolve
    {
      typedef  test_map <closed_hash_map <Key, Mapped, Hash, Equal,
                                          leak_test_allocator <allocator <pair <const Key,
                                                                                Mapped> > >,
                                          keep_hashes> >
               map_type;
    };

    static  const bool  linked = false;
  };

  struct Linked
  {
    template <typename Key, typename Mapped, typename Hash, typename Equal, bool keep_hashes>
    struct resolve
    {
      typedef  test_map <linked_hash_map <Key, Mapped, Hash, Equal,
                                          leak_test_allocator <allocator <pair <const Key,
                                                                                Mapped> > >,
                                          keep_hashes> >
               map_type;
    };

    static  const bool  linked = true;
  };


  template <typename Implementation,
            typename Key,
            typename Mapped,
            typename Hash  = MCT_HASH_NAMESPACE::hash <Key>,
            typename Equal = equal_to <Key>,
            bool _keep_hashes = false>
  struct Parameters
  {
    typedef  Key                                 key_type;
    typedef  Mapped                              mapped_type;
    typedef  pair <const key_type, mapped_type>  value_type;
    typedef  Hash                                hasher;
    typedef  Equal                               equal_to;

    static  const bool  keep_hashes = _keep_hashes;
    static  const bool  linked      = Implementation::linked;

    typedef  typename Implementation::template resolve <Key, Mapped, Hash, Equal,
                                                        _keep_hashes>::map_type
             map_type;
  };


  // To speed compilation up we list just a few possible combinations for each base type.
  typedef  mpl::vector <Parameters <Plain, int, int>,
                        Parameters <Linked, int, int>,
                        Parameters <Plain, int, string, int_hasher_1>,
                        Parameters <Plain, int, string,
                                    MCT_HASH_NAMESPACE::hash <int>, equal_to <int>, true>,
#                 if MCT_HAVE_LONG_LONG
                        Parameters <Plain, unsigned long long, string>,
                        Parameters <Plain, unsigned long long, int, unsigned_long_long_hasher>,
                        Parameters <Linked, unsigned long long, string,
                                    MCT_HASH_NAMESPACE::hash <unsigned long long>,
                                    equal_to <unsigned long long>, true>,
#                 endif
                        Parameters <Plain, string, int>,
                        Parameters <Plain, string, string, string_hasher_1>,
                        Parameters <Plain, string, int, string_hasher_1, equal_to <string>, true>,
                        Parameters <Linked, string, string>,
                        Parameters <Plain, foo, int>,
                        Parameters <Linked, foo, string>,
                        Parameters <Plain, bar, int>,
                        Parameters <Linked, bar, int> >
           test_parameters_std_equal;

  typedef  mpl::push_back <test_parameters_std_equal,
                           Parameters <Plain, int, string, int_hasher_special,
                                       int_comparator_special> >::type
           test_parameters;


#define COMMON_TEST_SETUP                                       \
  typedef  typename parameters::map_type      map_type;         \
  typedef  typename map_type::iterator        iterator;         \
  typedef  typename map_type::const_iterator  const_iterator;   \
  typedef  typename parameters::key_type      key_type;         \
  typedef  typename parameters::value_type    value_type;       \
  typedef  typename parameters::mapped_type   mapped_type;      \
  typedef  typename parameters::hasher        hasher;           \
  typedef  typename parameters::equal_to      equal_to;         \
  typedef  Data <key_type>                    key_data;         \
  typedef  Data <mapped_type>                 mapped_data;      \
  const bool  linked = parameters::linked


# include "tests/common-data.hpp"

}


BOOST_AUTO_TEST_SUITE (map_main)


BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map;
  foreach (const key_type& key, key_data::values1)
    map.insert (make_pair (key, mapped_type ()));
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_subscription_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map;
  foreach (const key_type& key, key_data::values1)
    map[key] = mapped_type ();
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_modification_through_iterator_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map;

  map.insert (make_pair (key_data::values1[0], mapped_type ()));
  map.find (key_data::values1[0])->second = mapped_data::values1[0];

  BOOST_CHECK_EQUAL (map.at (key_data::values1[0]), mapped_data::values1[0]);
}


BOOST_AUTO_TEST_SUITE_END ()


// Local variables:
// mode: c++
// c-basic-offset: 2
// indent-tabs-mode: nil
// fill-column: 90
// End:
