// Copyright (c) 2009, 2010 Paul Pogonyshev.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.


// Even though we perform our own integrity validation, and even comparisons to a known
// good implementation in our tests, we still let the tables validate themselves with
// MCT_SELF_VALIDATION.  If nothing else, this ensures that MCT_SELF_VALIDATION behaves
// correctly on valid tables.
#undef  MCT_SELF_VALIDATION
#define MCT_SELF_VALIDATION     1

#include "tests/set-common.hpp"
#include "tests/set-parameters.hpp"


namespace
{
# include "tests/common-data.hpp"
}


BOOST_AUTO_TEST_SUITE (set_exception_safety)


// Note that the tests do not look like checking almost anything.  That's because
// 'test_set' validates and compares itself with a known good implementation after each
// modification.  So, most actual checks just happen elsewhere.

BOOST_AUTO_TEST_CASE_TEMPLATE (test_range_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (ARRAY_RANGE (data::values1));
  if (linked)
    assert_identical_order (set, data::values1);
}


#if MCT_CXX0X_SUPPORTED

BOOST_AUTO_TEST_CASE_TEMPLATE (test_initializer_constructor_1,
                               implementation, test_implementations)
{
  typedef  typename implementation::template resolve <int>::set_type  set_type;

  set_type  set1 { 1, 2, 3, 4 };
  set_type  set2 { 1, 3, 1, 3, 2, 2, 2 };

  if (implementation::linked)
    {
      assert_identical_order (set1, vector <int> { 1, 2, 3, 4 });
      assert_identical_order (set2, vector <int> { 1, 3, 1, 3, 2, 2, 2 });
    }
}

#endif  // MCT_CXX0X_SUPPORTED


BOOST_AUTO_TEST_CASE_TEMPLATE (test_find_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (ARRAY_RANGE (data::values1));

  foreach (const type& value, data::values1)
    BOOST_CHECK (contains (set, value));

  foreach (const type& value, data::values2)
    {
      BOOST_CHECK_EQUAL (contains (set, value),
                         find (ARRAY_RANGE (data::values1), value) != ARRAY_END (data::values1));
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_count_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (ARRAY_RANGE (data::values1));

  foreach (const type& value, data::values1)
    BOOST_CHECK_EQUAL (set.count (value), 1u);

  foreach (const type& value, data::values2)
    {
      BOOST_CHECK_EQUAL (set.count (value),
                         (find (ARRAY_RANGE (data::values1), value) != ARRAY_END (data::values1)
                          ? 1u : 0u));
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_equal_range_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (ARRAY_RANGE (data::values1));

  foreach (const type& value, data::values1)
    {
      pair <const_iterator, const_iterator>  range (set.equal_range (value));
      BOOST_CHECK       (set.key_eq () (*range.first, value));
      BOOST_CHECK_EQUAL (++range.first, range.second);
    }

  foreach (const type& value, data::values2)
    {
      pair <const_iterator, const_iterator>  range (set.equal_range (value));

      if (find (ARRAY_RANGE (data::values1), value) == ARRAY_END (data::values1))
        BOOST_CHECK_EQUAL (range.first, range.second);
      else
        {
          BOOST_CHECK       (set.key_eq () (*range.first, value));
          BOOST_CHECK_EQUAL (++range.first, range.second);
        }
    }
}


#if MCT_CXX0X_SUPPORTED

BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_moving_1,
                               implementation, test_implementations)
{
  {
    typedef  typename implementation::template resolve <must_not_be_copied>::set_type  set_type;

    // Make it large, so it never rehashes and thus copies elements.
    set_type  set (1000);

    for (int k = 0; k < 100; ++k)
      {
        set.insert (must_not_be_copied (k));
        set.validate_integrity ();
      }
  }

  {
    typedef  typename implementation::template resolve <must_not_be_moved>::set_type  set_type;

    set_type  set (1000);

    for (int k = 0; k < 100; ++k)
      {
        must_not_be_moved  value (k);

        set.insert (value);
        set.validate_integrity ();
      }
  }
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_initializers_1,
                               implementation, test_implementations)
{
  typedef  typename implementation::template resolve <int>::set_type  set_type;

  set_type  set { 0, 1, 2 };
  set.insert ({ 3, 4, 5 });

  vector <int>         expected_vector { 0, 1, 2, 3, 4, 5 };
  unordered_set <int>  expected (expected_vector.begin (), expected_vector.end ());

  set.validate_integrity ();
  BOOST_CHECK_EQUAL (set, expected);

  if (implementation::linked)
    assert_identical_order (set, expected);
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_emplace_1,
                               implementation, test_implementations)
{
  typedef  typename implementation::template resolve <int_wrapper>::set_type  set_type;

  set_type  set;
  for (int k = 0; k < 50; ++k)
    {
      set.emplace (k);
      set.validate_integrity ();
    }

  BOOST_CHECK_EQUAL (set.size (), 50u);
  if (implementation::linked)
    assert_identical_order (set, create_vector (int_wrapper::wrap, set.size ()));
}

#endif  // MCT_CXX0X_SUPPORTED


BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_erase_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set;

  foreach (const type& value, data::values1)
    {
      // Simply insert and immediately erase one element.
      set.insert (value);
      BOOST_CHECK (set.erase (value));
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_comparison_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (ARRAY_RANGE (data::values1));
  set_type  set2 (ARRAY_RANGE (data::values1));
  set_type  set3 (ARRAY_RANGE (data::values2));
  set_type  set4 (ARRAY_RANGE (data::values2));

  BOOST_CHECK_EQUAL (set1, set2);
  BOOST_CHECK_EQUAL (set3, set4);
  BOOST_CHECK_NE    (set1, set3);
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_copy_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (ARRAY_RANGE (data::values1));
  set_type  set2 (ARRAY_RANGE (data::values2));
  set_type  set3 (set1);
  set_type  set4 (set2);

  BOOST_CHECK_EQUAL (set3, set1);
  BOOST_CHECK_EQUAL (set4, set2);

  if (linked)
    {
      assert_identical_order (set3, set1);
      assert_identical_order (set4, set2);
    }
}


#if MCT_CXX0X_SUPPORTED

BOOST_AUTO_TEST_CASE_TEMPLATE (test_move_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (move (set_type (ARRAY_RANGE (data::values1))));
  BOOST_CHECK_EQUAL (set, set_type (ARRAY_RANGE (data::values1)));
}

#endif  // MCT_CXX0X_SUPPORTED


BOOST_AUTO_TEST_CASE_TEMPLATE (test_assignment_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (ARRAY_RANGE (data::values1));
  set_type  set2;

  BOOST_CHECK_EQUAL (set2.size (), 0u);

  set2 = set1;
  BOOST_CHECK_EQUAL (set2, set1);

  if (linked)
    assert_identical_order (set1, set2);
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_self_assignment_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (ARRAY_RANGE (data::values1));
  set_type  set2 (ARRAY_RANGE (data::values1));

  BOOST_CHECK_EQUAL (set1, set2);
  if (linked)
    assert_identical_order (set1, set2);

  set1 = set1;

  BOOST_CHECK_EQUAL (set1, set2);
  if (linked)
    assert_identical_order (set1, set2);
}


#if MCT_CXX0X_SUPPORTED

BOOST_AUTO_TEST_CASE_TEMPLATE (test_move_assignment_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set;

  set = move (set_type (ARRAY_RANGE (data::values1)));
  BOOST_CHECK_EQUAL (set, set_type (ARRAY_RANGE (data::values1)));
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_initializer_assignment_1,
                               implementation, test_implementations)
{
  typedef  typename implementation::template resolve <int>::set_type  set_type;

  set_type  set { 1, 2, 3, 4 };
  set = { 4, 5, 6 };

  if (implementation::linked)
    assert_identical_order (set, vector <int> { 4, 5, 6 });
}

#endif  // MCT_CXX0X_SUPPORTED


BOOST_AUTO_TEST_CASE_TEMPLATE (test_iterators_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type        set (ARRAY_RANGE (data::values1));
  size_t          size = set.size ();
  const_iterator  scan = set.begin ();

  for (size_t k = 0; k != size; ++k, ++scan)
    {
      BOOST_CHECK (contains (set, *scan));

      if (k != 0)
        {
          const_iterator  copy (scan);
          --copy;
          ++copy;
          BOOST_CHECK_EQUAL (copy, scan);

          const_iterator  copy2 (scan);
          copy2--;
          copy2++;
          BOOST_CHECK_EQUAL (copy2, scan);
        }

      const_iterator  copy (scan);
      ++copy;
      --copy;
      BOOST_CHECK_EQUAL (copy, scan);

      const_iterator  copy2 (scan);
      copy2++;
      copy2--;
      BOOST_CHECK_EQUAL (copy2, scan);
    }

  BOOST_CHECK_EQUAL (scan, set.end ());
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_swap_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (ARRAY_RANGE (data::values1));
  set_type  set2 (ARRAY_RANGE (data::values2));
  set_type  set3 (ARRAY_RANGE (data::values1));
  set_type  set4 (ARRAY_RANGE (data::values2));

  BOOST_CHECK_EQUAL (set1, set3);
  BOOST_CHECK_EQUAL (set2, set4);

  if (linked)
    {
      assert_identical_order (set1, set3);
      assert_identical_order (set2, set4);
    }

  swap (set3, set4);

  BOOST_CHECK_EQUAL (set1, set4);
  BOOST_CHECK_EQUAL (set2, set3);

  if (linked)
    {
      assert_identical_order (set1, set4);
      assert_identical_order (set2, set3);
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_clear_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (ARRAY_RANGE (data::values1));
  set.clear ();
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_erase_value_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (ARRAY_RANGE (data::values1));
  set_type  set2 (ARRAY_BEGIN (data::values1) + 1, ARRAY_END (data::values1));

  // Make sure funny comparison functions don't ruin this test.
  if (set1 != set2)
    {
      BOOST_CHECK       (set1.erase (data::values1[0]));
      BOOST_CHECK_EQUAL (set1, set2);

      if (linked)
        assert_identical_order (set1, set2);
    }
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_erase_position_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (ARRAY_RANGE (data::values1));

  const size_t  num_passes_expected = set.size ();
  size_t        num_passes          = 0;

  for (iterator scan = set.begin (), end = set.end (); scan != end; ++num_passes)
    scan = set.erase (scan);

  BOOST_CHECK       (set.empty ());
  BOOST_CHECK_EQUAL (num_passes, num_passes_expected);
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_erase_range_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (ARRAY_RANGE (data::values1));

  set.erase (set.begin (), set.end ());
  BOOST_CHECK (set.empty ());
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_growth_1,
                               parameters, test_parameters_std_equal)
{
  COMMON_TEST_SETUP;

  // Note: fairly small, because testing after each step is slow.
  set_type  set;
  for (int k = 0; k < 1000; ++k)
    set.insert (data::generate (k));

  BOOST_CHECK_EQUAL (set.size (), 1000u);
  if (linked)
    assert_identical_order (set, create_vector (data::generate, set.size ()));
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_shrink_1,
                               parameters, test_parameters_std_equal)
{
  COMMON_TEST_SETUP;

  const size_t   num_items = 1000;
  vector <type>  items (create_vector (data::generate, num_items));
  set_type       set   (items.begin (), items.end ());

  BOOST_CHECK_EQUAL (set.size (), num_items);

  const_iterator  almost_end = set.end ();
  for (int k = 0; k < 10; ++k)
    --almost_end;

  set.erase (set.begin (), almost_end);

  BOOST_CHECK_EQUAL (set.size (), 10u);
  if (linked)
    assert_identical_order (set, create_vector (data::generate, num_items, num_items - 10));

  set.rehash (32);

  BOOST_CHECK_EQUAL (set.bucket_count (), 32u);
  if (linked)
    assert_identical_order (set, create_vector (data::generate, num_items, num_items - 10));
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_tiny_set_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  float  load_factors[] = { 0.01, 0.1, 0.25, 0.5, 0.75, 0.9, 0.99 };

  foreach (float load_factor, load_factors)
    {
      BOOST_TEST_CHECKPOINT ("testing with maximum load factor " << load_factor);

      set_type  set (1);
      set.max_load_factor (load_factor);

      for (int k = 0; k < 50; ++k)
        set.insert (data::generate (k));

      if (linked)
        assert_identical_order (set, create_vector (data::generate, 50));
    }
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_relink_1,
                               parameters, test_parameters_linked)
{
  COMMON_TEST_SETUP;

  vector <type>  initializers (create_vector (data::generate, 50));
  set_type       set (initializers.begin (), initializers.end ());

  assert_identical_order (set, initializers);

  set.relink (set.find (data::generate (40)), set.find (data::generate (20)));
  set.validate_integrity ();

  initializers.erase  (find (initializers.begin (), initializers.end (), data::generate (20)));
  initializers.insert (find (initializers.begin (), initializers.end (), data::generate (40)),
                       data::generate (20));

  assert_identical_order (set, initializers);

  // The following "moves" have no effect.

  set.relink (set.find (data::generate (5)), set.find (data::generate (5)));
  set.validate_integrity ();

  assert_identical_order (set, initializers);

  set.relink (set.find (data::generate (15)), set.find (data::generate (14)));
  set.validate_integrity ();

  assert_identical_order (set, initializers);
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_reverse_1,
                               parameters, test_parameters_linked)
{
  COMMON_TEST_SETUP;

  vector <type>  initializers (create_vector (data::generate, 50));
  set_type       set (initializers.begin (), initializers.end ());

  assert_identical_order (set, initializers);

  set.reverse ();
  set.validate_integrity ();
  reverse (initializers.begin (), initializers.end ());

  assert_identical_order (set, initializers);
}


BOOST_AUTO_TEST_SUITE_END ()


// Local variables:
// mode: c++
// c-basic-offset: 2
// indent-tabs-mode: nil
// fill-column: 90
// End:
